from langgraph.graph import StateGraph as StateGraphOrg
from langchain.load.dump import dumps
import json
import requests
from langchain_core.runnables import Runnable
from langchain_core.runnables.base import (
    RunnableLambda,
    RunnableLike,
    coerce_to_runnable,
)
from langgraph.graph.graph import END, START, CompiledGraph, Graph
from typing import Any, Callable, Dict, NamedTuple, Optional, Sequence

START = "__start__"
END = "__end__"

class StateGraphWork(StateGraphOrg):

    def __init__(self, options,serverUrl = ''):
        super().__init__(options)
        self.logList = []
        self.nodeList = []
        self.edgeList = []
        self.start = ''
        self.end = ''
        # self.serverUrl = options.get('server_url', None)
        self.serverUrl = serverUrl


    # 增加节点
    def add_node(self, key: str, action: RunnableLike) -> None:
        # nodelist中添加node节点的输入输出
        self.nodeList.append({'type': 'node', 'name': key})
        def on_end(run):
            log = {
                'type': 'node',
                'name': key,
                'input': run.inputs,
                'output': run.outputs,
            }
            self.logList.append(log)
        super().add_node(key,coerce_to_runnable(action).with_listeners(on_end=on_end))

    # 增加边
    def add_edge(self, start_key: str, end_key: str) -> None:
        super().add_edge(start_key,end_key)
        self.edgeList.append([start_key, end_key])


    def add_conditional_edges(
        self,
        start_key: str,
        condition: Callable[..., str],
        conditional_edge_mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        self.edgeList.append([start_key, conditional_edge_mapping])
        def wrapper_condition(state):
            # print("wrapper_condition")
            next_state = condition(state)
            if next_state == 'end':
                log = {
                    'nodes': self.nodeList,
                    'edges': self.edgeList,
                    'start': self.start,
                    'end': self.end,
                    'logs': self.logList,
                }
                # Save log to file
                print(dumps(log, sort_keys=True, indent=4, separators=(', ', ': ')))
                # Submit to server
                # if self.serverUrl:
                    # requests.post(self.serverUrl, data=log)
            else:
                self.logList.append({
                    'type': 'ConditionalEdges',
                    'name': start_key,
                    'input': state,
                    'output': next_state,
                })
            return next_state

        super().add_conditional_edges(start_key, wrapper_condition, conditional_edge_mapping)


    def setEntryPoint(self, agent):
        self.start = agent
        super().setEntryPoint(agent)

    def set_entry_point(self, key: str) -> None:
        super().set_entry_point(key)
        self.start = key

    def setFinishPoint(self, agent):
        self.end = agent

    def compile(self):
        return super().compile()

    def returnres(self):
        output = {
            'nodes': self.nodeList,
            'edges': self.edgeList,
            'logs': self.logList,
        }
        return output



import os
from dotenv import load_dotenv, dotenv_values

# 读环境变量
load_dotenv('.env')
env_configs = dotenv_values()


os.environ["OPENAI_API_KEY"] = "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"
os.environ["TAVILY_API_KEY"] = "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"

os.environ["LANGCHAIN_TRACING_V2"] = "true"
os.environ["LANGCHAIN_API_KEY"] = "ls__8b4e5a07430148a1b26d143ee40f6a59"


from langchain import hub
from langchain.agents import create_openai_functions_agent
from langchain_openai.chat_models import ChatOpenAI
from langchain_community.tools.tavily_search import TavilySearchResults

tools = [TavilySearchResults(max_results=1)]

# Get the prompt to use - you can modify this!
prompt = hub.pull("hwchase17/openai-functions-agent")

# Choose the LLM that will drive the agent
llm = ChatOpenAI(model="gpt-3.5-turbo-1106", streaming=True)

# Construct the OpenAI Functions agent
agent_runnable = create_openai_functions_agent(llm, tools, prompt)

from typing import TypedDict, Annotated, List, Union
from langchain_core.agents import AgentAction, AgentFinish
from langchain_core.messages import BaseMessage
import operator


class AgentState(TypedDict):
    # The input string
    input: str
    # The list of previous messages in the conversation
    chat_history: list[BaseMessage]
    # The outcome of a given call to the agent
    # Needs `None` as a valid type, since this is what this will start as
    agent_outcome: Union[AgentAction, AgentFinish, None]
    # List of actions and corresponding observations
    # Here we annotate this with `operator.add` to indicate that operations to
    # this state should be ADDED to the existing values (not overwrite it)
    intermediate_steps: Annotated[list[tuple[AgentAction, str]], operator.add]

from langchain_core.agents import AgentFinish
from langgraph.prebuilt.tool_executor import ToolExecutor

# This a helper class we have that is useful for running tools
# It takes in an agent action and calls that tool and returns the result
tool_executor = ToolExecutor(tools)


# Define the agent
def run_agent(data):
    agent_outcome = agent_runnable.invoke(data)
    return {"agent_outcome": agent_outcome}


# Define the function to execute tools
def execute_tools(data):
    # Get the most recent agent_outcome - this is the key added in the `agent` above
    agent_action = data["agent_outcome"]
    output = tool_executor.invoke(agent_action)
    return {"intermediate_steps": [(agent_action, str(output))]}


# Define logic that will be used to determine which conditional edge to go down
def should_continue(data):
    # If the agent outcome is an AgentFinish, then we return `exit` string
    # This will be used when setting up the graph to define the flow
    if isinstance(data["agent_outcome"], AgentFinish):
        return "end"
    # Otherwise, an AgentAction is returned
    # Here we return `continue` string
    # This will be used when setting up the graph to define the flow
    else:
        return "continue"

# from langgraph.graph import END, StateGraph
from StateGraphWork import END,StateGraphWork

# Define a new graph
workflow = StateGraphWork(AgentState)

# Define the two nodes we will cycle between
workflow.add_node("agent", run_agent)
workflow.add_node("action", execute_tools)

# Set the entrypoint as `agent`
# This means that this node is the first one called
workflow.set_entry_point("agent")

# We now add a conditional edge
workflow.add_conditional_edges(
    # First, we define the start node. We use `agent`.
    # This means these are the edges taken after the `agent` node is called.
    "agent",
    # Next, we pass in the function that will determine which node is called next.
    should_continue,
    # Finally we pass in a mapping.
    # The keys are strings, and the values are other nodes.
    # END is a special node marking that the graph should finish.
    # What will happen is we will call `should_continue`, and then the output of that
    # will be matched against the keys in this mapping.
    # Based on which one it matches, that node will then be called.
    {
        # If `tools`, then we call the tool node.
        "continue": "action",
        # Otherwise we finish.
        "end": END,
    },
)

# We now add a normal edge from `tools` to `agent`.
# This means that after `tools` is called, `agent` node is called next.
workflow.add_edge("action", "agent")

# Finally, we compile it!
# This compiles it into a LangChain Runnable,
# meaning you can use it as you would any other runnable
app = workflow.compile()

inputs = {"input": "what is the weather in sf", "chat_history": []}
for s in app.stream(inputs):
    # print(list(s.values())[0])
    print("----")


import os
from dotenv import load_dotenv, dotenv_values

# 读环境变量
load_dotenv('.env')
env_configs = dotenv_values()

os.environ["OPENAI_API_KEY"] = env_configs.get('OPENAI_API_KEY')
os.environ["TAVILY_API_KEY"] = env_configs.get('TAVILY_API_KEY')

os.environ["LANGCHAIN_TRACING_V2"] = "true"
os.environ["LANGCHAIN_API_KEY"] = env_configs.get('LANGCHAIN_API_KEY')

os.environ["http_proxy"] = env_configs.get('PROXY_URL')
os.environ["https_proxy"] = env_configs.get('PROXY_URL')
os.environ["OPENAI_PROXY"] = env_configs.get('PROXY_URL')

from langchain_community.tools.tavily_search import TavilySearchResults

tools = [TavilySearchResults(max_results=1)]

from langgraph.prebuilt import ToolExecutor

tool_executor = ToolExecutor(tools)

from langchain_openai import ChatOpenAI

# We will set streaming=True so that we can stream tokens
# See the streaming section for more information on this.
model = ChatOpenAI(temperature=0, streaming=True)

from langchain.tools.render import format_tool_to_openai_function

functions = [format_tool_to_openai_function(t) for t in tools]
model = model.bind_functions(functions)

from typing import TypedDict, Annotated, Sequence
import operator
from langchain_core.messages import BaseMessage


class AgentState(TypedDict):
    messages: Annotated[Sequence[BaseMessage], operator.add]

from langgraph.prebuilt import ToolInvocation
import json
from langchain_core.messages import FunctionMessage


# Define the function that determines whether to continue or not
def should_continue(state):
    messages = state["messages"]
    last_message = messages[-1]
    # If there is no function call, then we finish
    if "function_call" not in last_message.additional_kwargs:
        return "end"
    # Otherwise if there is, we continue
    else:
        return "continue"


# Define the function that calls the model
def call_model(state):
    messages = state["messages"]
    response = model.invoke(messages)
    # We return a list, because this will get added to the existing list
    return {"messages": [response]}


# Define the function to execute tools
def call_tool(state):
    messages = state["messages"]
    # Based on the continue condition
    # we know the last message involves a function call
    last_message = messages[-1]
    # We construct an ToolInvocation from the function_call
    action = ToolInvocation(
        tool=last_message.additional_kwargs["function_call"]["name"],
        tool_input=json.loads(
            last_message.additional_kwargs["function_call"]["arguments"]
        ),
    )
    # We call the tool_executor and get back a response
    response = tool_executor.invoke(action)
    # We use the response to create a FunctionMessage
    function_message = FunctionMessage(content=str(response), name=action.tool)
    # We return a list, because this will get added to the existing list
    return {"messages": [function_message]}

# from langgraph.graph import StateGraph, END
form xxxx import StateGraph,END

# Define a new graph
workflow = StateGraph(AgentState)

# Define the two nodes we will cycle between
workflow.add_node("agent", call_model)
workflow.add_node("action", call_tool)

# Set the entrypoint as `agent`
# This means that this node is the first one called
workflow.set_entry_point("agent")

# We now add a conditional edge
workflow.add_conditional_edges(
    # First, we define the start node. We use `agent`.
    # This means these are the edges taken after the `agent` node is called.
    "agent",
    # Next, we pass in the function that will determine which node is called next.
    should_continue,
    # Finally we pass in a mapping.
    # The keys are strings, and the values are other nodes.
    # END is a special node marking that the graph should finish.
    # What will happen is we will call `should_continue`, and then the output of that
    # will be matched against the keys in this mapping.
    # Based on which one it matches, that node will then be called.
    {
        # If `tools`, then we call the tool node.
        "continue": "action",
        # Otherwise we finish.
        "end": END,
    },
)

# We now add a normal edge from `tools` to `agent`.
# This means that after `tools` is called, `agent` node is called next.
workflow.add_edge("action", "agent")

# Finally, we compile it!
# This compiles it into a LangChain Runnable,
# meaning you can use it as you would any other runnable
app = workflow.compile()

from langchain_core.messages import HumanMessage

inputs = {"messages": [HumanMessage(content="what is the weather in sf")]}
app.invoke(inputs)